/*
 * Linux device driver programming Book(SB Creative corp)
 * Procfs sample program
 *
 * Update 2016/12/12 
 * (C) 2008 Yutaka Hirata
 */
#include <linux/module.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <net/net_namespace.h>

MODULE_LICENSE("Dual BSD/GPL");

/*
 * すでに /proc/driver/ は存在するが、その配下にサブディレクトリを
 * 作成して、/proc/driver/sample/sample_data エントリを作る。
 */
#define PARENT_DRV_PROCNAME NULL
#define DRV_PROCNAME "driver/sample"
#define PROCNAME "sample_data"

static struct proc_dir_entry *sample_proc = NULL;

static int as_show(struct seq_file *m, void *p)
{
	struct task_struct *tp = (struct task_struct *)p;

	printk("%p (%s)\n", tp, __func__);

	seq_printf(m, "[%s] pid=%d\n", tp->comm, tp->pid);
	seq_printf(m, "     tgid=%d\n", tp->tgid);
	seq_printf(m, "     state=%ld\n", tp->state);
	seq_printf(m, "     mm=0x%p\n", tp->mm);
	seq_printf(m, "     utime=%llu\n", tp->utime);
	seq_printf(m, "     stime=%llu\n", tp->stime);
	/* oomkilladj メンバは削除された。*/
	//seq_printf(m, "     oomkilladj=%d\n", tp->oomkilladj);
	seq_puts(m, "\n");

	return 0;
}

static int sample_proc_show(struct seq_file *m, void *v)
{
	struct task_struct *p = NULL;
	struct task_struct *tp;

	/* Acquire a read-only spinlock */
	rcu_read_lock();

	p = &init_task;   /* swapper process */

	for_each_process(p) {  /* init later process */
		as_show(m, p);
		tp = next_task(p);
		if (tp == &init_task) 
			break;
		p = tp;
	}

	/* Release a read-only spinlock */
	rcu_read_unlock();

    return 0;
}

static int sample_proc_open(struct inode *inode, struct file *file)
{
    return single_open(file, sample_proc_show, NULL);
}

static const struct file_operations sample_proc_fops = {
    .owner      = THIS_MODULE,
    .open       = sample_proc_open,
    .read       = seq_read,
    .llseek     = seq_lseek,
    .release    = single_release,
    .write      = NULL,
};

static int sample_init(void)
{
	struct proc_dir_entry *entry;

	/* /proc 配下にディレクトリを作る */
	sample_proc = proc_mkdir(DRV_PROCNAME, PARENT_DRV_PROCNAME);
	if (sample_proc == NULL) {
		printk(KERN_ERR "Unable to create `%s' proc directory", DRV_PROCNAME);
		goto error;
	}

	/* ディレクトリ配下にエントリを作る */
	entry = proc_create(PROCNAME, S_IRUGO | S_IWUSR, sample_proc, &sample_proc_fops);
	if (entry == NULL) {
		printk(KERN_WARNING "sample: unable to create /proc entry\n");
		goto error;
	}

	printk("driver loaded\n");
		
	return 0;

error:
	if (sample_proc) {
		remove_proc_entry(DRV_PROCNAME, PARENT_DRV_PROCNAME);
		sample_proc = NULL;
	}
	return -EIO;
}

static void sample_exit(void)
{
	if (sample_proc) {
		/* 不要になったエントリおよびディレクトリを削除する */
		remove_proc_entry(PROCNAME, sample_proc);
		remove_proc_entry(DRV_PROCNAME, PARENT_DRV_PROCNAME);
		sample_proc = NULL;
	}

	printk(KERN_ALERT "driver unloaded\n");
}

module_init(sample_init);
module_exit(sample_exit);

